#!/usr/bin/env python
# coding: utf-8

# Copyright (c) 2021 Mitsuru Ohno
#  Use of this source code is governed by a BSD-3-style
#  license that can be found in the LICENSE file.

# 08/02/2021, M. Ohno
# functions for MonomerClassifier and PolymerGenerator.

import numpy as np
import pandas as pd
from rdkit import rdBase, Chem
from rdkit.Chem import AllChem

def genmol(s):
    try:
        m = Chem.MolFromSmiles(s)
    except:
        m = np.nan
    return m


def gencSMI(m):
    try:
        cS = Chem.MolToSmiles(m)
    except:
        cS = np.nan
    return cS


#classify candidate compounds for mono-FG monomer
def monomer_sel_MFG(m, mons, excls):
    if pd.notna(m):
        chk = []
        if len(mons)!=0:
            for mon in mons:
                patt = Chem.MolFromSmarts(mon)
                if m.HasSubstructMatch(patt):
                    chk_excl=[]
                    for excl in excls:
                        excl_patt=Chem.MolFromSmarts(excl)
                        if m.HasSubstructMatch(excl_patt):
                            chk_excl.append(False)
                        else:
                            chk_excl.append(True)
                    if False in chk_excl:
                        chk.append(False)
                    else:
                        chk.append(True)
                else:
                    chk.append(False)
            if True in chk:
                fchk = True
            else:
                fchk = False
        else:
            fchk = False
    else:
        fchk = False
    return fchk


#classify candidate compounds for poly-FG monomer
#count objective FGs
def monomer_sel_PFG(m, mons, excls, minFG, maxFG):
    if pd.notna(m):
        chk_c = 0
        fchk_c = 0
        if len(mons)!=0:
            for mon in mons:
                patt = Chem.MolFromSmarts(mon)
                chk_c = len(m.GetSubstructMatches(patt))
                fchk_c = fchk_c + chk_c
            if minFG <= fchk_c <= maxFG:
                chk=[]
                for excl in excls:
                    excl_patt=Chem.MolFromSmarts(excl)
                    if m.HasSubstructMatch(excl_patt):
                        chk.append(False)
                    else:
                        chk.append(True)
                if False in chk:
                    fchk = False
                else:
                    fchk=True
            else:
                fchk = False
        else:
            fchk = (False)
    else:
        fchk = (False)
    return fchk


#define sequential polymerization for chain polymerization except polyolefine
def seq_chain(prod_P, targ_mon1, Ps_rxnL, mon_dic, monL):
    if Chem.MolToSmiles(prod_P) != '':
        if targ_mon1 not in ['vinyl', 'cOle']:
            seqFG2=Chem.MolFromSmarts(monL[[202][0]])
            seqFG3=Chem.MolFromSmarts(monL[[203][0]])
            seqFG4=Chem.MolFromSmarts(monL[[204][0]])
            while prod_P.HasSubstructMatch(seqFG2):
                prods = Ps_rxnL[202].RunReactants([prod_P])
                prod_P = prods[0][0]
                Chem.SanitizeMol(prod_P)
            while prod_P.HasSubstructMatch(seqFG3):
                prods = Ps_rxnL[203].RunReactants([prod_P])
                prod_P = prods[0][0]
                Chem.SanitizeMol(prod_P)
            while prod_P.HasSubstructMatch(seqFG4):
                prods = Ps_rxnL[204].RunReactants([prod_P])
                prod_P = prods[0][0]
                Chem.SanitizeMol(prod_P)
        else:
            prod_P=prod_P
    return prod_P


#define sequential polymerization for successive polymerization
def seq_successive(prod_P, targ_rxn, monL, Ps_rxnL, P_class):
    if Chem.MolToSmiles(prod_P) != '':
        seqFG0=Chem.MolFromSmarts(monL[[200][0]])
        seqFG1=Chem.MolFromSmarts(monL[[201][0]])
        seqFG2=Chem.MolFromSmarts(monL[[202][0]])
        seqFG3=Chem.MolFromSmarts(monL[[203][0]])
        seqFG4=Chem.MolFromSmarts(monL[[204][0]])
        seqFG5=Chem.MolFromSmarts(monL[[205][0]])
        seqFG6=Chem.MolFromSmarts(monL[[206][0]])
        if P_class not in ['polyolefin', 'polyoxazolidone', ]:
            while prod_P.HasSubstructMatch(seqFG1):
                prods = Ps_rxnL[201].RunReactants([prod_P])
                prod_P = prods[0][0]
                Chem.SanitizeMol(prod_P)
            while prod_P.HasSubstructMatch(seqFG2):
                prods = Ps_rxnL[202].RunReactants([prod_P])
                prod_P = prods[0][0]
                Chem.SanitizeMol(prod_P)
            while prod_P.HasSubstructMatch(seqFG3):
                prods = Ps_rxnL[203].RunReactants([prod_P])
                prod_P = prods[0][0]
                Chem.SanitizeMol(prod_P)
            while prod_P.HasSubstructMatch(seqFG4):
                prods = Ps_rxnL[204].RunReactants([prod_P])
                prod_P = prods[0][0]
                Chem.SanitizeMol(prod_P)
            while prod_P.HasSubstructMatch(seqFG5):
                prods = Ps_rxnL[205].RunReactants([prod_P])
                prod_P = prods[0][0]
                Chem.SanitizeMol(prod_P)
            while prod_P.HasSubstructMatch(seqFG6):
                if P_class =='polyimide':
                    prods = Ps_rxnL[207].RunReactants([prod_P])
                elif P_class =='polyester':
                    prods = Ps_rxnL[206].RunReactants([prod_P])
                prod_P = prods[0][0]
                Chem.SanitizeMol(prod_P)
        elif P_class in ['polyoxazolidone', ]:
            while prod_P.HasSubstructMatch(seqFG1):
                prods = Ps_rxnL[201].RunReactants([prod_P])
                prod_P = prods[0][0]
                Chem.SanitizeMol(prod_P)
            while prod_P.HasSubstructMatch(seqFG5):
                prods = Ps_rxnL[208].RunReactants([prod_P])
                prod_P = prods[0][0]
                Chem.SanitizeMol(prod_P)
        elif P_class in ['polyolefin', ]:
            while prod_P.HasSubstructMatch(seqFG0):
                prods = Ps_rxnL[200].RunReactants([prod_P])
                prod_P = prods[0][0]
                Chem.SanitizeMol(prod_P)
        else:
            prod_P=prod_P
    return prod_P


#homopolymerization
def homopolymR(mon1,mons,excls, targ_mon1, Ps_rxnL, mon_dic, monL):
    prod_P=mon1
    while monomer_sel_MFG(prod_P, mons, excls)== True: #生成したポリマーがさらに重合可能な場合、再度反応
        prods = Ps_rxnL[mon_dic[targ_mon1]].RunReactants([prod_P])
        try:
            prod_P = prods[0][0]
            Chem.SanitizeMol(prod_P)
            prod_P = seq_chain(prod_P, targ_mon1=targ_mon1, Ps_rxnL=Ps_rxnL, mon_dic=mon_dic, monL=monL)
        except:
            pass
    return [gencSMI(prod_P)] #20230904 revised returned Molobject to SMILES


#binarypolymerization
def bipolymR(reactant, targ_rxn, monL, Ps_rxnL, P_class):
    prod_P = Chem.MolFromSmiles('')
    prods = targ_rxn.RunReactants(reactant)
    try:
        prod_P = prods[0][0]
        Chem.SanitizeMol(prod_P)
        prod_P = seq_successive(prod_P, targ_rxn=targ_rxn, monL=monL, Ps_rxnL=Ps_rxnL, P_class=P_class)
    except:
        pass
    return [gencSMI(prod_P)] #20230904 revised returned Molobject to SMILES


#homopolymerization
def homopolymA(mon1,mons,excls, targ_mon1, Ps_rxnL, mon_dic, monL):
    prod_P=mon1
    while monomer_sel_MFG(prod_P, mons, excls)== True: #生成したポリマーがさらに重合可能な場合、再度反応
        prods = Ps_rxnL[mon_dic[targ_mon1]].RunReactants([prod_P])
        prod_Ps = []
        for prod_P in prods:
            try:
                Chem.SanitizeMol(prod_P[0])
                prod_P = prod_P[0]
                prod_P = seq_chain(prod_P, targ_mon1=targ_mon1, Ps_rxnL=Ps_rxnL, mon_dic=mon_dic, monL=monL)
                prod_Ps.append(prod_P)
            except:
                pass
    return [gencSMI(m) for m in prod_Ps]


#binarypolymerization
def bipolymA(reactant, targ_rxn, monL, Ps_rxnL, P_class):
    prod_P = Chem.MolFromSmiles('')
    prods = targ_rxn.RunReactants(reactant)
    prod_Ps = []
    for prod_P in prods:
        try:
            Chem.SanitizeMol(prod_P[0])
            prod_P = prod_P[0]
            prod_P = seq_successive(prod_P, targ_rxn=targ_rxn, monL=monL, Ps_rxnL=Ps_rxnL, P_class=P_class)
            prod_Ps.append(prod_P)
        except:
            pass
    return [gencSMI(m) for m in prod_Ps]
# end



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) 2021 Mitsuru Ohno
#  Use of this source code is governed by a BSD-3-style
#  license that can be found in the LICENSE file.

# polymer generator from classfied monomers.

import os
from pathlib import Path
import numpy as np
import pandas as pd
import json
import pickle
from rdkit import rdBase, Chem
from rdkit.Chem import AllChem, Draw
from .funclib import monomer_sel_MFG, monomer_sel_PFG, seq_chain, seq_successive, genmol, gencSMI

def biplym(df, targ = None, Pmode = None, dsp_rsl = None):
    if targ == None:
        targ = ['all', ]
    if Pmode == None:
        Pmode = 'a'
    if dsp_rsl == None:
        dsp_rsl = False


    db_file = os.path.join(str(Path(__file__).resolve().parent.parent), 'rules')
    with open(os.path.join(db_file, 'mon_dic.json'), 'r') as f:
        mon_dic = json.load(f)
    with open(os.path.join(db_file, 'mon_lst.json'), 'r') as f:
        monL = json.load(f)
    with open(os.path.join(db_file, 'excl_lst.json'), 'r') as f:
        exclL = json.load(f)
    with open(os.path.join(db_file, 'ps_rxn.pkl'), 'rb') as f:
        Ps_rxnL = pickle.load(f)
    with open(os.path.join(db_file, 'ps_class.json'), 'r') as f:
        Ps_classL = json.load(f)
    with open(os.path.join(db_file, 'ps_gen.pkl'), 'rb') as f:
        Ps_GenL = pickle.load(f)


    monL = {int(k): v for k, v in monL.items()}
    exclL = {int(k): v for k, v in exclL.items()}


    #set the generated polymer class
    targL = []
    if targ==['all', ]:
        targL = Ps_classL.keys()
    else:
        targL.extend(targ)
    for x in targL:
        if x not in Ps_classL.keys():
            print('oops! no such polymer class!')
            return
        else:
            pass

    #export fixed form DataFrame as pickle for next step
    DF =df.drop('ROMol', axis=1).dropna(subset=['smip_cand_mons'])
    DF_L = ['smip_cand_mons', ]
    for mon_class in mon_dic.keys():
        DF_L.append(mon_class)
    DF = DF[DF_L]
    DF_L = DF_L[1:]
    for col_nam in DF_L:
        DF[col_nam] = DF[col_nam].replace('False', '')
        DF[col_nam] = DF[col_nam].astype('bool')


    #select mode
    if Pmode == 'r':
        from .funclib import bipolymR, homopolymR
    elif Pmode == 'a':
        from .funclib import bipolymA, homopolymA
    else:
        raise Exception('invalid mode!')

    DF_Pgen = pd.DataFrame(columns=['mon1', 'mon2', 'polym', 'polymer_class'])

    #generate polymer
    for P_class in targL:
        for P_set in Ps_GenL[str(P_class)]:
            targ_mon1 = ''
            targ_mon2 = ''
            targ_mon1 = P_set[0]
            targ_mon2 = P_set[1]
            temp1 = []
            temp2 = []
            DF10 = DF[DF[targ_mon1]]
            temp1 = list(DF10['smip_cand_mons'])
            if len(temp1) != 0:
                if targ_mon2 != 'none':
                    DF20 = DF[DF[targ_mon2]]
                    temp2 = list(DF20['smip_cand_mons'])
                    del DF10, DF20
                    if len(temp2) != 0:
                        combs=[[m1, m2] for m1 in temp1 for m2 in temp2]
                        temp11=[]
                        temp21=[]
                        for comb in combs:
                            if comb[0]!=comb[1]:
                                temp11.append(comb[0])
                                temp21.append(comb[1])
                        DF_temp = pd.DataFrame()
                        DF_temp = pd.DataFrame(data={'mon1':temp11, 'mon2':temp21}, columns=['mon1', 'mon2'])
                        DF_temp['polymer_class'] = str(P_class)
                        targ_rxn=P_set[2]
                        if Pmode == 'r':
                            DF_temp['polym'] = DF_temp.apply(lambda x: [genmol(x['mon1']), genmol(x['mon2'])], axis=1).apply(bipolymR, targ_rxn=targ_rxn, monL=monL, Ps_rxnL=Ps_rxnL, P_class=P_class)
                            DF_Pgen = pd.concat([DF_Pgen, DF_temp], ignore_index=True, copy=False)
                        elif Pmode == 'a':
                            DF_temp['polym'] = DF_temp.apply(lambda x: [genmol(x['mon1']), genmol(x['mon2'])], axis=1).apply(bipolymA, targ_rxn=targ_rxn, monL=monL, Ps_rxnL=Ps_rxnL, P_class=P_class)
                            DF_Pgen = pd.concat([DF_Pgen, DF_temp], ignore_index=True, copy=False)
                else:
                    temp2 = ['' for i in range(len(temp1))]
                    DF_temp = pd.DataFrame()
                    DF_temp = pd.DataFrame(data={'mon1':temp1, 'mon2':temp2}, columns=['mon1', 'mon2'])
                    DF_temp['polymer_class'] = str(P_class)
                    mons=monL[mon_dic[targ_mon1]]
                    excls=exclL[mon_dic[targ_mon1]]
                    if Pmode == 'r':
                        DF_temp['polym'] = DF_temp.apply(lambda x: genmol(x['mon1']), axis=1).apply(homopolymR, mons=mons, excls=excls, targ_mon1=targ_mon1, Ps_rxnL=Ps_rxnL, mon_dic=mon_dic, monL=monL)
                        DF_Pgen = pd.concat([DF_Pgen, DF_temp], ignore_index=True, copy=False)
                    elif Pmode == 'a':
                        DF_temp['polym'] = DF_temp.apply(lambda x: genmol(x['mon1']), axis=1).apply(homopolymA, mons=mons, excls=excls, targ_mon1=targ_mon1, Ps_rxnL=Ps_rxnL, mon_dic=mon_dic, monL=monL)
                        DF_Pgen = pd.concat([DF_Pgen, DF_temp], ignore_index=True, copy=False)

    num_polym_react = len(DF_Pgen)
    DF_gendP = DF_Pgen.explode('polym')
    DF_gendP = DF_gendP.reset_index(drop=True)
    DF_gendP = DF_gendP.dropna(subset=['polym'])

    #adjust DataFrame
    DF_gendP['polym'].replace('', np.nan, inplace=True)

    #drpo duplicated polymerization reaction
    DF_gendP=DF_gendP.dropna(subset=['polym'])
    DF_gendP=DF_gendP[DF_gendP['mon1']!=DF_gendP['mon2']]
    DF_gendP['reactset']=np.sort(DF_gendP.loc[:,['mon1', 'mon2']].values).tolist()
    DF_gendP['reactset']=DF_gendP['reactset'].apply(set).apply(tuple)
    DF_gendP = DF_gendP.drop_duplicates(subset=['reactset', 'polym'])
    DF_gendP = DF_gendP.reset_index(drop=True)
    if dsp_rsl == True:
        if Pmode == 'a':
            print('run at advanced mode')
        elif Pmode == 'r':
            print('run at rapid mode')
        else:
            print('invalid mode')
        print('number of polymerization reactions = ', num_polym_react)
        print('number of generated polymers = ', len(DF_gendP))
    else:
        pass
    return DF_gendP
# #end



# Copyright (c) 2021 Mitsuru Ohno
#  Use of this source code is governed by a BSD-3-style
#  license that can be found in the LICENSE file.

# 08/21/2022, M. Ohno
# smipoly __init__

from ._version import __version__



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) 2021 Mitsuru Ohno
#  Use of this source code is governed by a BSD-3-style
#  license that can be found in the LICENSE file.

# 07/27/2021, M. Ohno
# smilesで与えた化合物リストを、モノマー別に分類する
# monomer categolization system of the compound list in SMILES
#
# Refernce:
# https://future-chem.com/rdkit-chemical-rxn/
# https://www.daylight.com/dayhtml_tutorials/languages/smarts/smarts_examples.html
# https://www.daylight.com/dayhtml/doc/theory/theory.smarts.html

import os
from pathlib import Path
import numpy as np
import pandas as pd
import json
from rdkit import rdBase, Chem
from rdkit.Chem import AllChem, Draw
from .funclib import genmol, gencSMI, monomer_sel_MFG, monomer_sel_PFG

def moncls(df, smiColn, minFG=None, maxFG=None, dsp_rsl=None):
    # #The default number of the samle class of FG were limited 2 to 4 in the same molecule for poly functionalized monomer.
    # (dataframe, smiColn, minFG = 2, maxFG = 4)
    if minFG == None:
        minFG = 2
    if maxFG == None:
        maxFG = 4
    if dsp_rsl == None:
        dsp_rsl = False


    #read source file
    DF01 = df
    smiColn = smiColn


    #append CO and HCHO for carbonate
    DF02 = DF01.copy()
    if smiColn in DF02.columns.to_list():
        #DF02 = DF02[['CID', smiColn]] #remove this row if not required
        DFadd = pd.DataFrame([['[C-]#[O+]'], ['C=O'], ], columns=[smiColn])
        DF02 = pd.concat([DF02, DFadd], ignore_index=True)
    else:
        print("invalid SMILES column name")


    #drop NA of smiles, and add chemical structure
    DF02['ROMol'] = DF02[smiColn].apply(genmol)
    DF02['smip_cand_mons'] = DF02['ROMol'].apply(gencSMI)


    db_file = os.path.join(str(Path(__file__).resolve().parent.parent), 'rules')
    with open(os.path.join(db_file, 'mon_vals.json'), 'r') as f:
        mon_vals = json.load(f) 
    with open(os.path.join(db_file, 'mon_dic.json'), 'r') as f:
        mon_dic = json.load(f)
    with open(os.path.join(db_file, 'mon_dic_inv.json'), 'r') as f:
        mon_dic_inv = json.load(f)
    with open(os.path.join(db_file, 'mon_lst.json'), 'r') as f:
        monL = json.load(f)
    with open(os.path.join(db_file, 'excl_lst.json'), 'r') as f:
        exclL = json.load(f)

    monL = {int(k): v for k, v in monL.items()}
    exclL = {int(k): v for k, v in exclL.items()}
    mon_dic_inv = {int(k): v for k, v in mon_dic_inv.items()}


    #classification for mono-functionalized monomer
    #count functional groupe, remove exclude compounds andjudge targetted monomer or not.
    for i in mon_vals[0]: 
        mons=()
        excls=()
        chk=[]
        chk_excl=[]
        mons = monL[i]
        excls = list(exclL[i])
        DF02[mon_dic_inv[i]] = DF02['ROMol'].apply(monomer_sel_MFG, mons=mons, excls=excls)
        if dsp_rsl==True:
            print(i)
            print(mon_dic_inv[i], ' = ', len(DF02[DF02[mon_dic_inv[i]]==True]), ' / ', len(DF02))
        else:
            pass


    #classification for poly-functionalized monomer
    for i in mon_vals[1]:
        mons=()
        excls=()
        mons=monL[i]
        excls=exclL[i]
        chk=[]
        DF02[mon_dic_inv[i]] = DF02['ROMol'].apply(monomer_sel_PFG, mons=mons, excls=excls, minFG=minFG, maxFG=maxFG)
        if dsp_rsl==True:
            print(i)
            print(mon_dic_inv[i], ' = ', len(DF02[DF02[mon_dic_inv[i]]==True]), ' / ', len(DF02))
        else:
            pass
    return DF02
#end
